/* Core Hardware: Cache management
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef CACHE_H
#define CACHE_H

#include <inttypes.h>

/** \file cache.h
 * \brief routines for manipulation the cache
 */

/** \brief enable the instruction cache
 * \details unlocks and invalidates the instruction cache and instruction tlb
 *  and activates the instruction cache
 * \note as a side-effect, the branch target buffer is also invalidated
 * \note look at the Intel XScale Core Manual on Page 4.3.2-52 (ICache) and
 *  7.2.2-83 (Control Register) */
void iCacheEnable();

/** \brief determine wether the instruction cache is activated
 * \return \c TRUE iff cache is enabled
 * \note look at the Intel XScale Core Manual on
 *  Page 7.2.2-83 (Control Register) */
bool iCacheIsEnabled();

/** \brief disable the instruction cache
 * \details disables the instruction cache, unlocks and invalidates the cache,
 *  branch target buffer and tlb
 * \note look at the Intel XScale Core Manual on Page 4.3.2-52 (ICache)
 *  and 7.2.2-83 (Control Register) */
void iCacheDisable();

/** \brief invalidate the instruction cache
 * \note as a side effect, the branch target buffer is also invalidated
 * \note this will not invalidate locked regions. If you need this, call
 *  iCacheUnlock() before
 * \note look at the XScale Core Manual on Page 4.3.3-53 (ICache) and
 *  7.2.8-78 (Cache Functions) */
void iCacheInvalidate();

/** \brief invalidate consecutive lines of the instruction cache
 * \param addr base address to be invalidated
 * \param len length of the area to be invalidated
 * \note This will not invalidate the branch target buffer. If you need this,
 *  call btbInvalidate() afterwards
 * \warning the given line should not be locked as this results in an unusable
 *  cache line
 * \note look at the XScale Core Manual on Page 7.2.8-78 (Cache Functions) */
void iCacheInvalidateLine(const void * addr, uint32_t len);

/** \brief lock instructions in the instruction cache
 * \param addr base address
 * \param length length
 * \note look at the XScale Core Manual on Page 4.3.4-54 (ICache) and
 *  7.2.10-90 (Cache Lock Down)
 * \note this code must be placed in a non-cacheable region as described in the
 *  Core Manual */
void iCacheLock(const void * addr, uint32_t len);

/** \brief unlock the instruction cache
 * \note look at the XScale Core Manual on Page 7.2.10-90 (Cache Lock Down) */
void iCacheUnlock();

/** \brief enable the branch target buffer
 * \details invalidates the buffer and activates it
 * \note look at the XScale Core Manual on Page 5.2.1-59 (BTB) and
 *  7.2.2-83 (Control Register) */
void btbEnable();

/** \brief determine wether the branch target buffer is enabled
 * \return \c TRUE iff the buffer is enabled
 * \note look at the XScale Core Manual on Page 7.2.2-83 (Control Register) */
bool btbIsEnabled();

/** \brief disable the branch target buffer
 * \details invalidates the buffer and deactivates it
 * \note look at the XScale Core Manual on Page 5.2.1-59 (BTB) and
 *  7.2.2-83 (Control Register) */
void btbDisable();

/** \brief invalidate the branch target buffer
 * \note look at the XScale Core Manual on Page 5.2.2-59 (BTB) and
 *  7.2.8-78 (Cache Functions) */
void btbInvalidate();

/** \brief enable the data cache.
 * \details unlocks and invalidates the data cache and data tlb,
 *  flushes the write buffer and activates the cache
 * \warning make sure that the MMU is enabled at this point
 * \note look at the XScale Core Manual on Page 6.3.2-69 (DCache) and
 *  7.2.2-83 (Control Register) */
void dCacheEnable();

/** \brief determine wether the data cache is activated
 * \return \c TRUE iff cache is enabled
 * \note look at the Intel XScale Core Manual on
 *  Page 7.2.2-83 (Control Register) */
bool dCacheIsEnabled();

/** \brief disable the data cache
 * \details flushes the write buffer, disables the data cache, unlocks and
 *  invalidates the cache and tlb
 * \note look at the Intel XScale Core Manual on Page 6-61 (DCache),
 *  7.2.2-83 (Control Register) */
void dCacheDisable();

/** \brief clean and invalidate the data and mini-data cache
 * \details cleans the data cache by allocating lines in it, then cleans the
 *  mini-data cache by loading abitrary data in it and then invalidates the
 *  caches
 * \note look at the Intel XScale Core Manual on Page 6.3.3.1-70 (DCache) */
void dCacheCleanAndInvalidate();

/** \brief invalidate consecutive lines of the data cache
 * \param addr base address to be invalidated
 * \param len length of the area to be invalidated
 * \warning This will not clean the cache. If you need this, call
 *  dCacheCleanLine() before
 * \warning the given line should not be locked as this results in an unusable
 *  cache line
 * \note look at the XScale Core Manual on Page 7.2.8-78 (Cache Functions) */
void dCacheInvalidateLine(const void * addr, uint32_t len);

/** \brief clean consecutive lines of the data cache
 * \param addr base address to be invalidated
 * \param len length of the area to be invalidated
 * \note look at the XScale Core Manual on Page 7.2.8-78 (Cache Functions) */
void dCacheCleanLine(const void * addr, uint32_t len);

/** \brief clean and invalidate consecutive lines of the data cache
 * \param addr base address to be cleaned and invalidated
 * \param len length of the area to be cleaned and invalidated
 * \note look at the XScale Core Manual on Page 7.2.8-78 (Cache Functions) */
void dCacheCleanAndInvalidateLine(const void * addr, uint32_t len);

/** \brief drain write buffer
 * \note look at the XScale Core Manual on Page 6.1.3-64 (DCache)
 *  and 7.2.8-78 (Cache Functions) */
void dCacheDrainBuffer();

#endif

