/* Core Hardware: MMU management
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

.include "core_internal.s"

/** \brief enable the Memory Management Unit (MMU)
 * \details disables the instruction cache, sets the translation table base,
 *  activates the mmu and reenables the instruction cache (if it was enabled
 *  before)
 * \param tbl translation table base
 * \note set the domain access control register first using mmuSetDomain()
 * \note look at the Intel XScale Core Manual on Page 3.4.1-42 (MMU) and
 *  7.2.2-83 (Control Register) */
.globl mmuEnable
mmuEnable:
.func mmuEnable
	/* save r4, r5 and lr */
	stmfd sp!, {r4, r5, lr}
	/* save r0 into r4 */
	mov r4, r0
	/* check if instruction cache is enabled */
	bl iCacheIsEnabled
	/* save result in r5 and update flags */
	movs r5, r0
	/* disable instruction cache if it is enabled */
	blne iCacheDisable

	/* set the mmu translation table base */
	mcr p15, 0, r4, c2, c0, 0
	/* wait for changes */
	CPWAIT r0
	
	/* activate the mmu by setting the M-Bit of the control register */
	SET_CONTROL_BITS r0, #CP15_CONTROL_M
	/* wait for changes */
	CPWAIT r0

	/* test if instruction cache was enabled */
	cmp r5, #0
	/* reenable instruction cache then */
	blne iCacheEnable

	/* restore r4, r5 and return */
	ldmfd sp!, {r4, r5, pc}
.endfunc

/** \brief determine wether the MMU is enabled
 * \return \c TRUE iff the MMU is enabled
 * \note look at the Intel XScale Core Manual on Page 7.2.2-83 (Control
 *  Register) */
.globl mmuIsEnabled
mmuIsEnabled:
.func mmuIsEnabled
	/* test I-Bit and save result (true or false) in r0 */
	TEST_CONTROL_BITS r0, #CP15_CONTROL_M
	/* return */
	mov pc, lr
.endfunc

/** \brief disable the Memory Management Unit (MMU)
 * \details disables the instruction cache, branch target buffer and data
 *  cache, deactivates the mmu and reenables the instruction cache (if it was
 *  enabled before)
 * \note look at the Intel XScale Core Manual on Page 3.4.1-42 (MMU) and
 *  7.2.2-83 (Control Register) */
.globl mmuDisable
mmuDisable:
.func mmuDisable
	/* save r4 and lr */
	stmfd sp!, {r4, lr}
	/* check if icache is enabled */
	bl iCacheIsEnabled
	/* store result in r4 and update flags */
	movs r4, r0
	/* disable instruction cache if it's enabled */
	blne iCacheDisable
	
	/* check if dcache is enabled */
	bl dCacheIsEnabled
	cmp r0, #0
	/* disable it if it's enabled */
	blne dCacheDisable
	
	/* check if btb is enabled */
	bl btbIsEnabled
	cmp r0, #0
	/* disable it if it's enabled */
	blne btbDisable
	
	/* unlock instruction tlb */
	ITLB_UNLOCK
	/* unlock data tlb */
	DTLB_UNLOCK
	
	/* deactivate MMU by unsetting M-Bit in control register */
	UNSET_CONTROL_BITS r0, #CP15_CONTROL_M
	/* wait for changes */
	CPWAIT r0
	
	/* invalidate tlb */
	TLB_INVALIDATE

	/* test if instruction cache was enabled before */
	cmp r4, #0
	/* reenable it, if it was enabled */
	blne iCacheEnable
	
	/* return */
	ldmfd sp!, {r4, pc}
.endfunc

/** \brief set the domain access control register
 * \param domain the domain to be set
 * \note look at the Intel XScale Core Manual on Page 7.2.4-85
 *  (Domain Access Control Register) */
.globl mmuSetDomain
mmuSetDomain:
.func mmuSetDomain
	/* set domain access control register */
	mcr p15, 0, r0, c3, c0, 0
	/* wait for change */
	CPWAIT r0
	/* return */
	mov pc, lr
.endfunc

/** \brief get the domain access control register
 * \return value of the domain access control register
 * \note look at the Intel XScale Core Manual on Page 7.2.4-85
 *  (Domain Access Control Register) */
.globl mmuGetDomain
mmuGetDomain:
.func mmuGetDomain
	/* get domain access control register */
	mrc p15, 0, r0, c3, c0, 0
	/* return */
	mov pc, lr
.endfunc

/** \brief invalidate instruction tlb
 * \note this won't affect locked entries. use iTlbUnlock() if needed
 * \note look at the Intel XScale Core Manual on Page 3.4.1-42 (MMU) and
 *  7.2.9-89 (TLB Operations) */
.globl iTlbInvalidate
iTlbInvalidate:
.func iTlbInvalidate
	/* invalidate instruction tlb */
	ITLB_INVALIDATE
	/* return */
	mov pc, lr
.endfunc

/** \brief invalidate data tlb
 * \note this won't affect locked entries. use dTlbUnlock() if needed
 * \note look at the Intel XScale Core Manual on Page 3.4.1-42 (MMU) and
 *  7.2.9-89 (TLB Operations) */
.globl dTlbInvalidate
dTlbInvalidate:
.func dTlbInvalidate
	/* invalidate data tlb */
	DTLB_INVALIDATE
	/* return */
	mov pc, lr
.endfunc

/** \brief lock an entry in the instruction tlb
 * \param addr VMA of entry to be locked
 * \note look at the Intel XScale Core Manual on Page 3.4.3-43 (MMU) and
 *  7.2.11-91 (TLB Lock Down) */
.globl iTlbLockLine
iTlbLockLine:
.func iTlbLockLine
	/* translate and lock instruction tlb entry */
	mcr p15, 0, r0, c10, c4, 0
	/* return */
	mov pc, lr
.endfunc

/** \brief lock an entry in the data tlb
 * \param addr VMA of entry to be locked
 * \note look at the Intel XScale Core Manual on Page 3.4.3-43 (MMU) and
 *  7.2.11-91 (TLB Lock Down) */
.globl dTlbLockLine
dTlbLockLine:
.func dTlbLockLine
	/* translate and lock data tlb entry */
	mcr p15, 0, r0, c10, c8, 0
	/* return */
	mov pc, lr
.endfunc

/** \brief unlock instruction tlb
 * \note look at the Intel XScale Core Manual on Page 3.4.3-43 (MMU) and
 *  7.2.11-91 (TLB Lock Down) */
.globl iTlbUnlock
iTlbUnlock:
.func iTlbUnlock
	/* unlock instruction tlb */
	ITLB_UNLOCK
	/* return */
	mov pc, lr
.endfunc

/** \brief unlock data tlb
 * \note look at the Intel XScale Core Manual on Page 3.4.3-44 (MMU) and
 *  7.2.11-91 (TLB Lock Down) */
.globl dTlbUnlock
dTlbUnlock:
.func iTlbUnlock
	/* unlock data tlb */
	DTLB_UNLOCK
	/* return */
	mov pc, lr
.endfunc

