/* Debug functions
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <os_arm_bare_registers_def.h>
#include <stdio.h>
#include <stdarg.h>
#include "debug.h"
#include "config.h"

/** \defgroup debug Debug Interface
 * @{
 */

/** \file debug.c
 * \brief debug interface and instrumentation functions
 */

static void init(struct Driver * driver)
	__attribute__((no_instrument_function));

/** \brief debug driver structure */
struct Driver debugDriver = {
	.name = "Debug",
	
	.init = &init
};

/** \brief send data over stuart
 * \param msg message
 * \note \a no_instrument_function to prevent loops in instrumentation
 * \bug can fail if the 64'th character or a multiple is a newline or
 *  the end of the buffer to be sent */
__attribute__((no_instrument_function))
static inline void send(const char * msg) {
	char c;
	uint8_t sent = 0;
	
	/* flush buffer */
	while (!(STLSR & LSR_TEMT));
	
	/* send message */
	while(1) {
		c = *msg++; /* read one character and advance to next position */
		
		/* append newline at end */
		if (!c) {
			STTHR = '\r';
			STTHR = '\n';
			break;
		}
		
		/* extend newline by carriage return */
		if (c == '\n') {
			STTHR = '\r';
			++sent;
		}
		
		/* write one byte */
		STTHR = c;
		++sent;
		
		if (sent >= 64) {
			sent = 0;
			/* flush buffer */
			while (!(STLSR & LSR_TEMT));
		}
	}
}

#ifdef DEBUG_TRACE
/** \brief called by every function on enter
 * 
 * sends the address over the debug interface
 *
 * \param addr address
 * \param callsite callsite
 * \note \a no_instrument_function to prevent loops in instrumentation */
__attribute__((no_instrument_function))
void __cyg_profile_func_enter(void * addr, void * callsite) {
	/* 'e' for "Enter" and some buffer */
	static char buf[] = "edeadbeef";
	/* put address in message */
	sprintf(buf + 1, "%08x", (uint32_t)addr);
	/* send */
	send(buf);
}

/** \brief called by every function on leave
 * 
 * sends the address over the debug interface
 *
 * \param addr address
 * \param callsite callsite
 * \note \a no_instrument_function to prevent loops in instrumentation */
__attribute__((no_instrument_function))
void __cyg_profile_func_exit(void * addr, void * callsite) {
	/* 'l' for "Leave" and some buffer */
	static char buf[] = "ldeadbeef";
	/* put address in message */
	sprintf(buf + 1, "%08x", (unsigned int)addr);
	/* send */
	send(buf);
}
#endif

/** \brief initialize debug interface
 * \note \a no_instrument_function as the interface is not initialized
 *  on entering that function */
__attribute__((no_instrument_function))
static void init(struct Driver * driver) {
	/* stuart */
	GPIO_SET_ALT_FUNC(46,2,GPIO_IN);
	GPIO_SET_ALT_FUNC(47,1,GPIO_OUT);
	
	/* If sensorboard is used */
	GPIO_SET_ALT_FUNC(81,0,GPIO_OUT)
	GPCR(81) |= GPIO_BIT(81);
	
	/* enable clock */
	CKEN |= CKEN_CKEN5;
	
	STLCR |= LCR_DLAB;  // set uart to initial Baudrate 115200 bps
	STDLL = 0x8;
	STDLH = 0;
	STLCR &= ~(LCR_DLAB);
	// set uart to 8 data bits transmitted in each character
	STLCR = LCR_WLS(3);
	
	/* enable fifo */
	STFCR = FCR_TRFIFOE;
	
	/* enable stuart */
	STIER |= IER_UUE;
	
	/* send begin message */
	send("\nBEGIN");
}

/** \brief send a formatted debug message
 * \param format format
 * \param ... arguments
 * \return bytes actually passed
 * \note the buffer size is limited to #DEBUG_BUFFER_SIZE
 * \note \a no_instrument_function to prevent loops in instrumentation */
__attribute__((no_instrument_function, format(printf,1,2)))
int debugf(const char * format, ...) {
	char buf[DEBUG_BUFFER_SIZE];
	va_list ap;
	int ret;
	
	va_start(ap, format);
	ret = vsnprintf(buf, sizeof(buf), format, ap);
	
	debug(buf);
	
	return ret;
}

/** \brief send a debug message
 * \param msg message
 * \note \a no_instrument_function to prevent loops in instrumentation */
__attribute__((no_instrument_function))
void debug(const char * msg) {
	/* send 'm' as message mark, it is flushed on entering send() */
	STTHR = 'm';
	send(msg);
}

/** @} */

