/* Functions for overwriting the original bootloader
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <string.h>
#include <inttypes.h>
#include <os_arm_bare_registers_def.h>
#include <os_arm_bare_interrupt.h>
#include "debug.h"
#include "flash.h"
#include "dmalloc.h"
#include "led.h"

static const uint32_t vectors[] = {
	0xe59ff018, 0xe59ff018, 0xe59ff018, 0xe59ff018,
	0xe59ff018, 0xe59ff018, 0xe59ff018, 0xe59ff018
};

const int BOOT_SECTOR_SIZE = 32 * 1024;

extern const uint32_t bootvectors;

void deploy() {
	uint8_t bootsector[BOOT_SECTOR_SIZE];
	
	/* compare flash vectors with the backup vectors */
	if (!memcmp(&bootvectors, vectors, sizeof(vectors))) {
		debugf("Deploy: nothing to do");
		return;
	}
	
	ledOn(LED_BLUE);
	ledOn(LED_RED);
	ledOff(LED_GREEN);
	
	debugf("Deploy: installing bootloader by overwriting the vectors");
	
	/* disable interrupts as the interrupt redirection is overwritten by the
		erase */
	SENF_DISABLE_INTERRUPTS();
	
	/* build new bootsector */
	memcpy(bootsector, &bootvectors, BOOT_SECTOR_SIZE);
	memcpy(bootsector, vectors, sizeof(vectors));
	
	/* unlock boot block */
	flashBlockUnlock(0x0);
	if (flashBlockQueryLock(0x0)) {
		debug("Deploy: Warning: unlocking seems to be failed");
	}
	
	/* erase boot block */
	if (!flashBlockErase(0x0)) {
		debug("Deploy: Erase failed");
		ledOff(LED_BLUE);
		while (TRUE);
	}
	
	uint32_t pos;
	for (pos = 0; pos < BOOT_SECTOR_SIZE; pos += 64) {
		/* write 64-byte wise */
		if (!flashBlockWriteBuffered(pos,
			(uint16_t*)(bootsector + pos), 32)) {
			debugf("Deploy: could not write at position 0x%08x", pos);
			ledOff(LED_BLUE);
			while (TRUE);
		}
	}
	
	debug("Deploy: installation done");
	
	/* enable interrupts and execute */
	ledOff(LED_BLUE);
	ledOff(LED_RED);
	SENF_ENABLE_INTERRUPTS();
}

