/* Debug code for memory allocations
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef DEBUG_MALLOC_H
#define DEBUG_MALLOC_H

#include <malloc.h>
#include "debug.h"

/** \addtogroup debug
 * @{
 */

/** \file dmalloc.h
 * \brief debug malloc and friends
 */

#ifdef DEBUG_MEMORY
/** \brief dump some memory usage info using the debug interface */
static inline void memoryInfo() {
	struct mallinfo m = mallinfo();
	debugf("Mem: total reserved: %d, free %d, in use: %d",
		m.arena, m.fordblks, m.uordblks);
}
#else
/** \brief dummy macro to ignore calls of memoryInfo if debug is disabled */
#define memoryInfo()
#endif

#ifdef DEBUG_MEMORY

/** \brief allocate memory and print statistics */
#define dmalloc(size) ({ \
	void * ptr; \
	memoryInfo(); \
	ptr = malloc(size); \
	debugf("malloc from %s (%s:%d) with size=%lu returned ptr=%p", \
		__FUNCTION__, __FILE__, __LINE__, (size_t)(size), ptr); \
	memoryInfo();\
	ptr; \
})

/** \brief allocate aligned memory and print statistics */
#define dmemalign(alignment, size) ({ \
	void * ptr; \
	memoryInfo(); \
	ptr = memalign(alignment, size); \
	debugf("memalign from %s (%s:%d) with alignment=%lu and size=%lu " \
		"returned ptr=%p", \
		__FUNCTION__, __FILE__, __LINE__, (size_t)(alignment), (size_t)(size), \
		ptr); \
	memoryInfo();\
	ptr; \
})

/** \brief free memory and print statistics */
#define dfree(ptr) ({ \
	memoryInfo(); \
	free(ptr); \
	debugf("free from %s (%s:%d) with ptr=%p", \
		__FUNCTION__, __FILE__, __LINE__, ptr); \
	memoryInfo(); \
})

/** \brief reallocate memory and print statistics */
#define drealloc(ptr, size) ({ \
	void * nptr; \
	memoryInfo(); \
	nptr = realloc((ptr), (size)); \
	debugf("realloc from %s (%s:%d) with ptr=%p and size=%lu " \
		"returned nptr=%p", \
		__FUNCTION__, __FILE__, __LINE__, ptr, (size_t)(size), nptr); \
	memoryInfo(); \
	nptr; \
})

#else

/** \brief allocate memory (wrapper) */
#define dmalloc(size) malloc(size)
/** \brief allocate aligned memory (wrapper) */
#define dmemalign(alignment, size) memalign(alignment, size)
/** \brief free memory (wrapper) */
#define dfree(ptr) free(ptr)
/** \brief reallocate memory (wrapper) */
#define drealloc(ptr, size) realloc(ptr, size)

#endif

/** @} */

#endif

