/* Driver framework
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "driver.h"

/** \file driver.c
 * \brief Generic Driver routines
 */

/** Initiailize a driver.
 * The init routine of a driver should initialize the subsystem
 *  (e.g. setup registers) but not start it
 * \param driver the driver */
__attribute__((no_instrument_function))
void driverInit(struct Driver * driver) {
#ifdef DRIVER_DEBUG
	if (driver->isInit)
		debugf("Warning: Driver %s is already initialized", driver->name);
#endif
	if (driver->init)
		driver->init(driver);
	
	driver->isInit = TRUE;
}

/** Start a driver.
 * The start routine of a driver should start the subsystem (e.g. enable the
 *  appropriate interrupts)
 * \param driver the driver */
__attribute__((no_instrument_function))
void driverStart(struct Driver * driver) {
#ifdef DRIVER_DEBUG
	if (driver->isStart)
		debugf("Warning: Driver %s is already started", driver->name);
#endif
	
	if (!driver->isInit)
		driverInit(driver);
	
	if (driver->start)
		driver->start(driver);
	
	driver->isStart = TRUE;
}

/** Stop a driver.
 * The stop routine of a driver should stop the subsystem (e.g. disable the
 *  appropriate interrupts) but not destroy it, such that the start routine
 *  can reenable the subsystem properly.
 * \param driver the driver */
void driverStop(struct Driver * driver) {
#ifdef DRIVER_DEBUG
	if (!driver->isStart)
		debugf("Warning: Driver %s is not started", driver->name);
#endif
	if (driver->stop)
		driver->stop(driver);
	
	driver->isStart = FALSE;
}

/** Cleanup a driver.
 * The Cleanup routine of a driver should revert the effects of the
 *  init call (e.g. free some memory).
 * \param driver the driver */
void driverFinish(struct Driver * driver) {
#ifdef DRIVER_DEBUG
	if (!driver->isInit)
		debugf("Warning: Driver %s is not initialized", driver->name);
#endif
	
	if (driver->isStart)
		driver->stop(driver);
	
	if (driver->finish)
		driver->finish(driver);
	
	driver->isInit = FALSE;
}

