/* USB Driver: Dynamic Configuration
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include <os_arm_bare_registers_def.h>
#include "usb_main_internal.h"
#include "usb_descriptors.h"

/** \defgroup usbconfig USB Configuration
 * \ingroup usb
 * @{
 */

/** \file usb_configuration.c
 * \brief USB Configuration
 */

/** \cond INTERNAL */

/** \brief register a string
 *
 * maps a string to the next available index
 
 * \param string the string to be registered */
static void registerString(const char * string) {
	/* register only non-empty strings and if the map is not full */
	if (string != NULL && string[0] != '\0' &&
		usbUdc.stringsUsed < USB_CONFIG_MAX_STRINGS) {
		usbUdc.descriptorStrings[usbUdc.stringsUsed++] = string;
	}
}

/** \brief get the index of a previously registered string
 * \note operates on pointers, not on string contents
 * \param string the string
 * \return index or \c 0 iff string was not registered before */
uint8_t usbConfigGetStringIndex(const char * string) {
	uint8_t idx;
	
	/* iterate over strings */
	for (idx = 0; idx < usbUdc.stringsUsed; ++idx) {
		if (usbUdc.descriptorStrings[idx] == string) {
			/* return the index biased by one as the string index counts
				from zero in the USB Protocol */
			return idx + 1;
		}
	}
	
	return 0;
}

/** \brief get a registered string by its index.
 * \param index index of the string
 * \return the string or \c NULL iff the index does not exist */
const char * usbConfigGetString(uint8_t index) {
	if (index == 0 || index > usbUdc.stringsUsed)
		return NULL;
	
	return usbUdc.descriptorStrings[index - 1];
}

/** \endcond */

/** \brief load a new usb configuartion
 * \param devConf device configuration */
void usbConfigLoad(struct UsbDevice * devConf) {
	uint8_t confIdx, ifaceIdx, settingIdx;
	
	/* set configuration */
	usbUdc.config = devConf;
	
	/* reset string map */
	usbUdc.stringsUsed = 0;
	
	/* register device strings */
	registerString(devConf->devManufacturer);
	registerString(devConf->devProduct);
	registerString(devConf->devSerialNumber);
	
	for (confIdx = 0; confIdx < devConf->numConfigurations; ++confIdx) {
		const struct UsbConfiguration * conf = devConf->configurations[confIdx];
		
		/* register configuration description string */
		registerString(conf->confDescription);
	
		for (ifaceIdx = 0; ifaceIdx < conf->numInterfaces; ++ifaceIdx) {
			const struct UsbInterface * iface = conf->interfaces[ifaceIdx];
			
			for (settingIdx = 0; settingIdx < iface->numAltSettings;
				++settingIdx) {
				const struct UsbAltSetting * setting =
					iface->altSettings[settingIdx];
			
				/* register interface description string */
				registerString(setting->altDescription);
			}
		}
	}
}

/** \brief get the full configuration length of a configuration
 * \param conf configuration
 * \return size of the configuration descriptor plus sizes of all
 * 	interface descriptors and their endpoint descriptors  */
uint16_t usbConfigGetConfigurationLength(const struct UsbConfiguration * conf) {
	uint16_t length;
	uint8_t ifaceIdx, settingIdx;
	
	/* size of the configuration descriptor */
	length = sizeof(struct ConfigurationDescriptor);
	
	for (ifaceIdx = 0; ifaceIdx < conf->numInterfaces; ++ifaceIdx) {
		/* for each interface in that configuration */
		const struct UsbInterface * iface = conf->interfaces[ifaceIdx];
		
		for (settingIdx = 0; settingIdx < iface->numAltSettings; ++settingIdx) {
			/* for each alternate setting in the interface */
			const struct UsbAltSetting * setting = iface->altSettings[settingIdx];
			
			/* size of the interface descriptor */
			length += sizeof(struct InterfaceDescriptor);
		
			/* for each endpoint in that setting: size of endpoint descriptor */
			length += setting->numEndpoints * sizeof(struct EndpointDescriptor);
		}
	}
	
	return length;
}

/** \cond INTERNAL
 * \brief load the endpoint registers from the current device configuration */
void usbConfigDeploy() {
	uint8_t confIdx, ifaceIdx, settingIdx, endpointIdx, udcEndpointIdx;
	
	/* configure endpoint zero */
	UDCCSR0 = 0;
	
	/* begin at endpoint index 1 (zero is special) */
	udcEndpointIdx = 1;
	
	/* iterate over configurations */
	for (confIdx = 0; confIdx < usbUdc.config->numConfigurations; ++confIdx) {
		struct UsbConfiguration * conf = usbUdc.config->configurations[confIdx];
		
		/* iterate over interfaces */
		for (ifaceIdx = 0; ifaceIdx < conf->numInterfaces; ++ifaceIdx) {
			struct UsbInterface * iface = conf->interfaces[ifaceIdx];
			
			/* iterate over alternate settings */
			for (settingIdx = 0; settingIdx < iface->numAltSettings;
				++settingIdx) {
				struct UsbAltSetting * setting = iface->altSettings[settingIdx];
				
				/* iterate over endpoints */
				for (endpointIdx = 0; endpointIdx < setting->numEndpoints;
					++endpointIdx) {
					struct UsbEndpoint * endpoint =
						setting->endpoints[endpointIdx];
					
					/* configure endpoint register */
					UDCCRAX(udcEndpointIdx) =
						/* configuration */
						UDCCRAX_CN(confIdx + 1) |
						/* interface */
						UDCCRAX_IN(ifaceIdx) |
						/* alternate setting */
						UDCCRAX_AISN(settingIdx) |
						/* endpoint address */
						UDCCRAX_EP(endpoint->epAddress) |
						/* endpoint type */
						(endpoint->epType == USB_ENDPOINT_TYPE_ISOCHRONOUS ? 
							UDCCRAX_ET_ISOCHRONOUS :
							(endpoint->epType == USB_ENDPOINT_TYPE_BULK ?
							UDCCRAX_ET_BULK : UDCCRAX_ET_INTERRUPT)) |
						/* endpoint direction */
						(endpoint->epDirection == USB_ENDPOINT_DIR_OUT ?
							UDCCRAX_ED_OUT : UDCCRAX_ED_IN) |
						/* maximum packet size */
						UDCCRAX_MPS(endpoint->epMaxPacketSize) |
						/* enable double buffering if configured */
						(endpoint->doubleBuffer ? UDCCRAX_DE : 0 )|
						/* enable */
						UDCCRAX_EE;

					/* endpoint mapping */
					usbUdc.ep[udcEndpointIdx] = endpoint;
					endpoint->udcEndpoint = udcEndpointIdx;
					
					/* rest of endpoint is initialized when the bus is
					 * configured */
					
					++udcEndpointIdx;
				}
			}
		}
	}
	
	/* number of endpoints used */
	usbUdc.endpointsUsed = udcEndpointIdx;
	
	/* disable _all_ remaining endpoints */
	for (; udcEndpointIdx < 24; ++udcEndpointIdx) {
		UDCCRAX(udcEndpointIdx) = 0;
		if (udcEndpointIdx < USB_CONFIG_NUM_ENDPOINTS)
			usbUdc.ep[udcEndpointIdx] = NULL;
	}
}
/** \endcond */

/** @} */

