/* USB Driver: Definition of USB Descriptors
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#ifndef USB_DESCRIPTORS_H
#define USB_DESCRIPTORS_H

#include <inttypes.h>
#include "usb_configuration_internal.h"

/** \cond INTERNAL */

/** \addtogroup usbcomm
 * @{
 */

/** \file usb_descriptors.h
 * \brief USB Descriptor structures
 *
 * Contains all structures, which are filled from the data of the
 *  configuration interface in usbConfigLoad()
 */

/* USB Request Classes */
/** \brief USB Request Type
 * \param reqtype bmRequestType field of ::DeviceRequest
 * \return request class */
#define USB_REQUEST_TYPE(reqtype) (((reqtype) >> 5) & 3)
/** \brief Standard Request Type */
#define USB_REQUEST_TYPE_STANDARD	0x0
/** \brief Class Request Type */
#define USB_REQUEST_TYPE_CLASS		0x1
/** \brief Vendor Request Type */
#define USB_REQUEST_TYPE_VENDOR		0x2
/** \brief Reserved Request Type */
#define USB_REQUEST_TYPE_RESERVED	0x3

/* USB Standard Request Types */
/** \brief GET STATUS request */
#define USB_GET_STATUS			0x0
/** \brief CLEAR FEATURE request */
#define USB_CLEAR_FEATURE		0x1
/** \brief SET FEATURE request */
#define USB_SET_FEATURE			0x3
/** \brief SET ADDRESS request */
#define USB_SET_ADDRESS			0x5
/** \brief GET DESCRIPTOR request */
#define USB_GET_DESCRIPTOR		0x6
/** \brief SET DESCRIPTOR request */
#define USB_SET_DESCRIPTOR		0x7
/** \brief GET CONFIGURATION request */
#define USB_GET_CONFIGURATION	0x8
/** \brief SET CONFIGURATION request */
#define USB_SET_CONFIGURATION	0x9
/** \brief GET INTERFACE request */
#define USB_GET_INTERFACE		0xA
/** \brief SET INTERFACE request */
#define USB_SET_INTERFACE		0xB
/** \brief SYNCH FRAME request */
#define USB_SYNCH_FRAME			0xC

/* USB Descriptor Types */
/** \brief Device Descriptor Type 
 * \see DeviceDescriptor */
#define USB_DESCRIPTOR_DEVICE						0x1
/** \brief Configuration Descriptor Type
 * \see ConfigurationDescriptor */
#define USB_DESCRIPTOR_CONFIGURATION				0x2
/** \brief String Descriptor Type
 * \see StringDescriptor
 * \see StringDescriptor0 */
#define USB_DESCRIPTOR_STRING						0x3
/** \brief Interface Descriptor Type
 * \see InterfaceDescriptor */
#define USB_DESCRIPTOR_INTERFACE					0x4
/** \brief Endpoint Descriptor Type
 * \see EndpointDescriptor */
#define USB_DESCRIPTOR_ENDPOINT						0x5
/** \brief Device Qualifier Descriptor Type */
#define USB_DESCRIPTOR_DEVICE_QUALIFIER				0x6
/** \brief Other Speed Configuration Descriptor Type */
#define USB_DESCRIPTOR_OTHER_SPEED_CONFIGURATION	0x7
/** \brief Interface Power Descriptor Type */
#define USB_DESCRIPTOR_INTERFACE_POWER				0x8

/* USB Configuration Descriptor Attributes */
/** \brief Configuration Descriptor Bus Powered Attribute (USB 1.0)
 * \see ConfigurationDescriptor */
#define USB_CONFIGURATION_BUS_POWERED_10 (1 << 7)
/** \brief Configuration Descriptor Bus Powered Attribute (USB 1.1)
 * \see ConfigurationDescriptor */
#define USB_CONFIGURATION_BUS_POWERED_11 (1 << 6)
/** \brief Configuration Descriptor Bus Remote Wakeup Attribute
 * \see ConfigurationDescriptor */
#define USB_CONFIGURATION_REMOTE_WAKEUP (1 << 5)

/** \brief USB Bus Power
 * \param p power in \c mA
 * \return bus power value for request
 * \see ConfigurationDescriptor */
#define USB_POWER(p) ((p) > 500 ? 250 : ((p) >> 1) + ((p) & 0x1))

/* USB Endpoint Descriptor Attributes */
/** \brief Endpoint Descriptor OUT direction attribute
 * \see EndpointDescriptor */
#define USB_ENDPOINT_OUT (0 << 7)
/** \brief Endpoint Descriptor IN direction attribute
 * \see EndpointDescriptor */
#define USB_ENDPOINT_IN (1 << 7)
/** \brief Endpoint Descriptor Endpoint number
 * \param n endpoint number
 * \see EndpointDescriptor */
#define USB_ENDPOINT(n) (((n) & 0xf) << 0)
/** \brief Endpoint Descriptor Control Transfer attribute
 * \see EndpointDescriptor */
#define USB_ENDPOINT_CONTROL (0 << 0)
/** \brief Endpoint Descriptor Isochronous Transfer attribute
 * \see EndpointDescriptor */
#define USB_ENDPOINT_ISOCHRONOUS (1 << 0)
/** \brief Endpoint Descriptor Bulk Transfer attribute
 * \see EndpointDescriptor */
#define USB_ENDPOINT_BULK (2 << 0)
/** \brief Endpoint Descriptor Interrupt Transfer attribute
 * \see EndpointDescriptor */
#define USB_ENDPOINT_INTERRUPT (3 << 0)
/** \brief Endpoint Descriptor Max Packet size
 * \param n max packet size
 * \see EndpointDescriptor */
#define USB_ENDPOINT_PACKET_SIZE(n) ((n) & 0x7ff)

/** \brief structure of a device request as defined in USB 1.1 spec */
struct DeviceRequest {
	/** \brief bitmap request type */
	uint8_t bmRequestType;
	/** \brief request */
	uint8_t bRequest;
	union {
		/** \brief request value */
		uint16_t wValue;
		struct {
			/** \brief descriptor index */
			uint8_t bIndex;
			/** \brief descriptor type */
			uint8_t bType;
		} __attribute__ ((packed)) descriptor;
	} __attribute__((packed));
	/** \brief request index */
	uint16_t wIndex;
	/** \brief request length */
	uint16_t wLength;
} __attribute__((packed));

/** \brief structure of a device descriptor as defined in USB 1.1 spec */
struct DeviceDescriptor {
	/** \brief descriptor length */
	uint8_t bLength;
	/** \brief descriptor type */
	uint8_t bDescriptorType;
	/** \brief usb version in \c BCD
	 * \note 0x0110 for USB 1.1 */
	uint16_t bcdUSB;
	/** \brief device Class */
	uint8_t bDeviceClass;
	/** \brief device SubClass */
	uint8_t bDeviceSubClass;
	/** \brief device Protocol */
	uint8_t bDeviceProtocol;
	/** \brief max packet size on endpoint zero */
	uint8_t bMaxPacketSize0;
	/** \brief vendor id */
	uint16_t idVendor;
	/** \brief product id */
	uint16_t idProduct;
	/** \brief device version in \c BCD */
	uint16_t bcdDevice;
	/** \brief string index to manufacturer string */
	uint8_t iManufacturer;
	/** \brief string index to product string */
	uint8_t iProduct;
	/** \brief string index to serial number string */
	uint8_t iSerialNumber;
	/** \brief number of configurations for this device */
	uint8_t bNumConfigurations;
} __attribute__((packed));

/** \brief structure of an endpoint descriptor as defined in USB 1.1 spec */
struct EndpointDescriptor {
	/** \brief descriptor length */
	uint8_t bLength;
	/** \brief descriptor type */
	uint8_t bDescriptorType;
	/** \brief endpoint address */
	uint8_t bEndpointAddress;
	/** \brief endpoint attributes */
	uint8_t bmAttributes;
	/** \brief max packet size */
	uint16_t wMaxPacketSize;
	/** \brief interval
	 * \note used for isochronous and interrupt transfers only */
	uint8_t bInterval;
} __attribute__ ((packed));

/** \brief structure of an interface descriptor as defined in USB 1.1 spec */
struct InterfaceDescriptor {
	/** \brief descriptor length */
	uint8_t bLength;
	/** \brief descriptor type */
	uint8_t bDescriptorType;
	/** \brief interface number */
	uint8_t bInterfaceNumber;
	/** \brief alt setting number */
	uint8_t bAlternateSetting;
	/** \brief number of endpoints */
	uint8_t bNumEndpoints;
	/** \brief interface Class */
	uint8_t bInterfaceClass;
	/** \brief interface SubClass */
	uint8_t bInterfaceSubClass;
	/** \brief interface Protocol */
	uint8_t bInterfaceProtocol;
	/** \brief string index to interface description */
	uint8_t iInterface;
} __attribute__ ((packed));

/** \brief structure of a configuration descriptor as defined in USB 1.1 spec */
struct ConfigurationDescriptor {
	/** \brief descriptor length */
	uint8_t bLength;
	/** \brief descriptor type */
	uint8_t bDescriptorType;
	/** \brief total length of all descriptors
	 * \note the configuration descriptor appends all interface descriptors
	 *  with all their endpoint descriptors.
	 *  This field contains the length of all those descriptors */
	uint16_t wTotalLength;
	/** \brief number of interfaces */
	uint8_t bNumInterfaces;
	/** \brief configuration number */
	uint8_t bConfigurationValue;
	/** \brief string index to configuration description */
	uint8_t iConfiguration;
	/** \brief bitmap with configuration attributes */
	uint8_t bmAttributes;
	/** \brief max power description in \c mA/2
	 * \see USB_POWER */
	uint8_t bMaxPower;
} __attribute__ ((packed));

/** \brief structure of a string descriptor as defined in USB 1.1 spec */
struct StringDescriptor {
	/** \brief descriptor length */
	uint8_t bLength;
	/** \brief descriptor type */
	uint8_t bDescriptorType;
	/** \brief string data */
	uint16_t bString[];
} __attribute__ ((packed));

/** \brief String descriptor zero is special as it encodes language id(s) */
struct StringDescriptor0 {
	/** \brief descriptor length */
	uint8_t bLength;
	/** \brief descriptor type */
	uint8_t bDescriptorType;
	/** \brief language ids */
	uint16_t wLANGID[];
} __attribute__ ((packed));

/** @} */

/** \endcond */

#endif

