/* USB Driver: Endpoint handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include <inttypes.h>
#include <os_arm_bare_registers_def.h>
#include "usb_main_internal.h"
#include "usb_endpoint_internal.h"
#include "usb_request_internal.h"
#include "usb_debug.h"
#include "../common.h"
#include "../dmalloc.h"
#include "../core/cache.h"

/** \cond INTERNAL */

/** \defgroup usbcomm USB Communication
 * @{
 */

/** \file usb_endpoint.c
 * \brief USB Endpoint handling
 */

#if 1
/** \brief Short Packet and DMA Enable flag must be set in the UDCCSRx register
 * by the dma controller iff a short packet is sent
 * \note aligned that it can be used by DMA controller directly
 */
static const aligned32 uint32_t dmaDataSP = UDCCSRAX_SP | UDCCSRAX_DME;

static void sentDataDma(uint8_t channel, uint32_t remainingLength,
	void * endpoint);
static void receivedDataDma(uint8_t channel, uint32_t remainingLength,
	void * endpoint);
#endif

static void endpointFinish(struct UsbEndpoint * ep, enum UsbRequestStatus status);
static void endpointInit(struct UsbEndpoint * ep);
static void endpointFlush(struct UsbEndpoint * ep);

/** \brief called by interrupt handler
 *
 * if a interrupt on one of the endpoints occured */
void usbEndpointInterrupt() {
	uint8_t udcEndpointIdx;
	uint32_t udcisr0 = UDCISR0 & UDCISR0_EP_MASK;
	uint32_t udcisr1 = UDCISR1 & UDCISR1_EP_MASK;
	uint32_t udcisr;
	
	if (udcisr0 & UDCISR_IR) {
		/* interrupt on EP0 occured, ack and handle it */
		UDCISR0 = UDC_INT(0, UDCISR_IR);
		usbControlPipeInterrupt(!!(udcisr0 & UDCISR_IR_PC),
			!!(udcisr0 & UDCISR_IR_FE));
	}
	
	/* begin at endpoint 1 */
	udcisr = udcisr0 >> 2;
	for (udcEndpointIdx = 1; udcEndpointIdx < usbUdc.endpointsUsed;
		++udcEndpointIdx) {
		if (udcisr & UDCISR_IR) {
			/* interrupt on endpoint endpointIdx */
			usbEPDebug(usbUdc.ep[udcEndpointIdx], "Interrupt");
			
			/* ack and handle it */
			UDCISR(udcEndpointIdx) = UDC_INT(udcEndpointIdx, UDCISR_IR);
			if (usbUdc.ep[udcEndpointIdx]->useDma) {
				struct UsbEndpoint * ep = usbUdc.ep[udcEndpointIdx];
				/* this will cause a DMA interrupt which delivers the data */
				dmaStopChannel(ep->dmaChannel);
			} else {
				usbEndpointHandleData(usbUdc.ep[udcEndpointIdx]);
			}
		}
		
		udcisr >>= 2;
		
		if (udcEndpointIdx == 15) {
			/* switch register */
			udcisr = udcisr1;
		}
	}
}

/** \brief nuke endpoint (dequeue all requests)
 * \param ep endpoint to be nuked
 * \param status status for the requests to complete */
void usbEndpointNuke(struct UsbEndpoint * ep, enum UsbRequestStatus status) {
	usbEPInfo(ep, "Nuking endpoint with status %d (has req: %s)", status,
		ep->req ? "yes" : "no");
	
	while (ep->req)
		usbRequestComplete(ep, ep->req, status);
}

/** \brief Configure communication system when USB Configuration changes
 * \param newAltsetting new alternate setting, from which endpoints are derived
 * \note called by busReconfigInterrupt() */
void usbEndpointsConfigure(struct UsbAltSetting * newAltsetting) {
	uint8_t endpointIdx;
	
	/* unconfigure all endpoints */
	usbEndpointsFinishAll(USB_REQUEST_STATUS_RECONFIGURE);
	
	/* configure new endpoints */
	for (endpointIdx = 0; endpointIdx < newAltsetting->numEndpoints; ++endpointIdx)
		endpointInit(newAltsetting->endpoints[endpointIdx]);
}

/** \brief finish all endpoints
 * nukes all requests on all endpoints
 * \param status status to complete requests */
void usbEndpointsFinishAll(enum UsbRequestStatus status) {
	uint8_t endpointIdx;
	struct UsbAltSetting * setting = usbUdc.curAltSetting;
	
	if (setting)
		for (endpointIdx = 0; endpointIdx < setting->numEndpoints; ++endpointIdx)
			endpointFinish(setting->endpoints[endpointIdx], status);
}

/** \brief finish an endpoint
 * nukes all requests on an endpoint
 * \param ep endpoint to be finished
 * \param status status to complete requests */
static void endpointFinish(struct UsbEndpoint * ep, enum UsbRequestStatus status) {
	/* disable interrupt, nuke requests and flush endpoint */
	UDCICR(ep->udcEndpoint) &= ~UDC_INT(ep->udcEndpoint, UDCICR_IE);
	
	if (ep->useDma) {
		dmaReleaseChannel(ep->dmaChannel);
		
		/* do not use dmaFreeDescriptorChain here, because the descriptors
		 * may not be chained */
		if (ep->dmaDataDesc) {
			dfree(ep->dmaDataDesc);
			ep->dmaDataDesc = NULL;
		}
		if (ep->dmaEndDesc) {
			dfree(ep->dmaEndDesc);
			ep->dmaEndDesc = NULL;
		}
	}
	
	usbEndpointNuke(ep, status);
	endpointFlush(ep);
}

/** \brief initialize an endpoint
 * \param ep endpoint to be initialized */
static void endpointInit(struct UsbEndpoint * ep) {
	ep->req = NULL;
	ep->enabled = TRUE;
	ep->inHandle = FALSE;
	ep->dmaRunning = FALSE;
	
	endpointFlush(ep);
	
	ep->dmaDataDesc = NULL;
	ep->dmaEndDesc = NULL;
	
	if (ep->useDma) {
		usbEPDebug(ep, "DMA Endpoint, enabling dma");
		
		if (ep->epDirection == USB_ENDPOINT_DIR_IN) {
			ep->dmaDataDesc = dmaAllocateDescriptor();
			ep->dmaEndDesc = dmaAllocateDescriptor();
			
			/* descriptor for setting SP and DME flags in UDCCSR register
			 * after sending a short packet */
			dmaSetupDescMemToMem(ep->dmaEndDesc, TRUE, &dmaDataSP,
				(void*)&UDCCSRAX(ep->udcEndpoint), 3, sizeof(dmaDataSP));
			
			/* allocate dma channel */
			if (!dmaGetChannel(0, &ep->dmaChannel, 24 + ep->udcEndpoint,
				&sentDataDma, (void*)ep)) {
				usbEPErr(ep, "Could not get a dma Channel");
				return; /* TODO */;
			}
		} else {
			/* allocate dma channel */
			if (!dmaGetChannel(0, &ep->dmaChannel, 24 + ep->udcEndpoint,
				&receivedDataDma, (void*)ep)) {
				usbEPErr(ep, "Could not get a dma Channel");
				return; /* TODO */;
			}
		}
		
		usbWrite_UDCCSR(ep, UDCCSRAX_DME);
	}
}

/** \brief flush an endpoint
 * \param ep endpoint to be flushed */
static void endpointFlush(struct UsbEndpoint * ep) {
	volatile register uint32_t * byteReg;
	uint32_t bogus;
	
	/* do nothing on control pipe */
	if (usbIsControlPipe(ep))
		return;
	
	byteReg = &UDCDR(ep->udcEndpoint);
	
	if (ep->epDirection == USB_ENDPOINT_DIR_OUT) {
		while (!usbEndpointIsEmpty(ep))
			bogus = *byteReg;
	} else {
		usbWrite_UDCCSR(ep, UDCCSRAX_PC | UDCCSRAX_FEF | UDCCSRAX_TRN |
			(ep->epType == USB_ENDPOINT_TYPE_INTERRUPT ? 0 : UDCCSRAX_SST) |
			(ep->useDma ? UDCCSRAX_DME : 0));
	}
}

static void receivedDataDma(uint8_t channel, uint32_t remainingLength,
	void * endpoint) {
	struct UsbEndpoint * ep = (struct UsbEndpoint*)endpoint;
	struct UsbRequest * req = ep->req;
	
	ASSERT(req != NULL);
	
	ep->dmaRunning = FALSE;
	
	req->actual = req->length - remainingLength;
	
	/* invalidate data cache after dma transfer */
	dCacheInvalidateLine(req->data, req->actual);
	prefetch(req->data);
	
	usbEPDebug(ep, "Got DMA data on with length %d", req->actual);
	
	/* ack data and enable dma */
	if (UDCCSRAX(ep->udcEndpoint) & UDCCSRAX_SP &&
		!(UDCCSRAX(ep->udcEndpoint) & UDCCSRAX_FS))
		usbWrite_UDCCSR(ep, UDCCSRAX_DME | UDCCSRAX_PC);
	
	/* complete request */
	usbRequestComplete(ep, req, USB_REQUEST_STATUS_SUCCESS);
	
	/* handle next request if there is one */
	usbHandleDmaRequest(ep);
}

/** \brief DMA callback after data transfer is done for an \c IN endpoint
 *
 * free data if requested and setup next request if there are more data to send
 *
 * \param channel dma channel
 * \param remainingLength unused
 * \param endpoint endpoint */
static void sentDataDma(uint8_t channel, uint32_t remainingLength,
	void * endpoint) {
	struct UsbEndpoint * ep = (struct UsbEndpoint*)endpoint;
	struct UsbRequest * req = ep->req;
	
	ASSERT(req != NULL);
	
	ep->dmaRunning = FALSE;
	
	/* ack request and enable dma again */
	usbWrite_UDCCSR(ep, UDCCSRAX_DME | UDCCSRAX_PC);
	
	/* complete request */
	usbRequestComplete(ep, req, USB_REQUEST_STATUS_SUCCESS);
	
	/* handle next request if there is one */
	usbHandleDmaRequest(ep);
}

/** @} */

/** \endcond */

