/* USB Driver: Endpoint data handling
   Copyright (c) 2010 Networked Systems Group, University of Kaiserslautern

This file is part of the Imote2 Bootloader

The driver contains ideas taken from the linux gadget device driver,
licenced under the GPL.

The Bootloader is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

The Bootloader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include <stdlib.h>
#include <inttypes.h>
#include <os_arm_bare_registers_def.h>
#include "usb_configuration_internal.h"
#include "usb_endpoint_internal.h"
#include "usb_request.h"
#include "usb_request_internal.h"
#include "usb_main_internal.h"
#include "usb_debug.h"
#include "../assert.h"
#include "../common.h"
#include "../dma.h"

/** \cond INTERNAL */

/** \ingroup usbcomm
 * @{
 */

/** \file usb_endpoint_data.c
 * \brief USB Endpoint handling: Interrupt communication
 */

/** \brief write a packet from a request into an endpoint
 * \param ep endpoint to be written into
 * \param req request to be written
 * \return number of bytes written */
uint32_t usbEndpointWritePacket(struct UsbEndpoint * ep,
	struct UsbRequest * req) {
	uint32_t len, count, remain;
	int i;
	uint32_t * buf;
	uint8_t * buf8;
	volatile register uint32_t * byteReg;
	
	/* get the data to be sent */
	buf = (uint32_t*)(req->data + req->actual);
	prefetch(buf);
	
	len = min(req->length - req->actual, ep->epMaxPacketSize);
	req->actual += len;
	
	byteReg = &UDCDR(ep->udcEndpoint);
	
	remain = len & 0x3;
	count = len & ~(0x3);
	/* write words */
	for (i = count; i > 0; i -= 4)
		*byteReg = *buf++;
	
	buf8 = (uint8_t*)buf;
	/* write (maybe remaining) bytes */
	for (i = remain; i > 0; i--)
		_PXAREG8(byteReg) = *buf8++;
	
	usbEPDebug(ep, "wrote %d+%d bytes", count, remain);
	
	return len;
}

/** \brief read a packet from an endpoint into a request
 * \param ep endpoint to be read from
 * \param req request to be written into
 * \return number of bytes read */
inline uint32_t usbEndpointReadPacket(struct UsbEndpoint * ep,
	struct UsbRequest * req) {
	uint32_t bytesLeft, bufSpace, count;
	int i;
	uint32_t * buf;
	volatile register uint32_t * byteReg;
	
	bytesLeft = usbEndpointBytesRemain(ep);
	bufSpace = req->length - req->actual;
	
	buf = (uint32_t*)(req->data + req->actual);
	prefetch(buf);
	
	byteReg = &UDCDR(ep->udcEndpoint);
	
	if (!usbEndpointIsEmpty(ep))
		count = min(bytesLeft, bufSpace);
	else
		count = 0;
	
	for (i = count; i > 0; i -= 4)
		*buf++ = *byteReg;
	req->actual += count;
	
	usbWrite_UDCCSR(ep, UDCCSRAX_PC);
	
	usbEPDebug(ep, "read %d bytes", count);
	
	return count;
}

/** \brief read as much as possible from an endpoint into a request
 * \param ep endpoint to be read from
 * \param req request to be written in
 * \return true if request was filled */
static bool readFifo(struct UsbEndpoint * ep, struct UsbRequest * req) {
	uint32_t count, isShort;
	bool completed = FALSE;
	
	while (usbEndpointHasPkt(ep)) {
		count = usbEndpointReadPacket(ep, req);
		isShort = (count < ep->epMaxPacketSize);
		usbEPDebug(ep, "read count: %d bytes%s req %p %d/%d", count,
			isShort ? "/S" : "", req, req->actual, req->length);
		/* completion */
		if (isShort || req->actual == req->length) {
			completed = TRUE;
			break;
		}
	}
	return completed;
}

/** \brief write as much as possible from a request into an endpoint
 * \param ep endpoint to be written into
 * \param req request to be read from
 * \return true if request was fully transmitted */
static bool writeFifo(struct UsbEndpoint * ep, struct UsbRequest * req) {
	uint32_t count, total = 0, udccsr;
	bool isLast = FALSE, isShort, completed = FALSE;
	
	do {
		isShort = FALSE;
		
		udccsr = UDCCSRAX(ep->udcEndpoint);
		if (udccsr & UDCCSRAX_PC) {
			debugf("Clearing Transmit Complete on ep %d", ep->udcEndpoint);
			usbWrite_UDCCSR(ep, UDCCSRAX_PC);
		}
		if (udccsr & UDCCSRAX_TRN) {
			debugf("Clearing Underrun on on ep %d", ep->udcEndpoint);
			usbWrite_UDCCSR(ep, UDCCSRAX_TRN);
		}
		
		count = usbEndpointWritePacket(ep, req);
		total += count;
		
		if (count < ep->epMaxPacketSize) {
			isLast = TRUE;
			isShort = TRUE;
		} else {
			isLast = !(req->length > req->actual);
		}
		
		if (isShort)
			usbWrite_UDCCSR(ep, UDCCSRAX_SP);
		
		if (isLast) {
			completed = 1;
			break;
		}
	} while(!usbEndpointIsFull(ep));
	
	usbEPDebug(ep, "wrote count: %d bytes%s%s, left: %d req=%p",
		total, isLast ? "/L" : "", isShort ? "/S" : "",
		req->length -req->actual, req);
	
	return completed;
}

/** \brief handle data on an endpoint
 * \param ep endpoint
 * \note called by usbEndpointInterrupt() */
void usbEndpointHandleData(struct UsbEndpoint * ep) {
	struct UsbRequest * req;
	bool completed;
	uint32_t udccsr;
	
	/* detect recursions */
	if (ep->inHandle)
		return;
	
	ep->inHandle = TRUE;
	
	do {
		completed = FALSE;
		
		udccsr = UDCCSRAX(ep->udcEndpoint);
		req = ep->req;
		
		if (udccsr & (UDCCSRAX_SST | UDCCSRAX_TRN))
			UDCCSRAX(ep->udcEndpoint) = udccsr & (UDCCSRAX_SST | UDCCSRAX_TRN);
		
		if (!req)
			break;
		
		if (ep->epDirection == USB_ENDPOINT_DIR_IN) {
			if (!usbEndpointIsFull(ep))
				completed = writeFifo(ep, req);
		} else {
			if (usbEndpointHasPkt(ep))
				completed = readFifo(ep, req);
		}
		
		if (completed)
			usbRequestComplete(ep, req, USB_REQUEST_STATUS_SUCCESS);
	} while (completed);
	
	ep->inHandle = FALSE;
}

/** @} */

/** \endcond */


